﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Drawing;
using System.Text;
using System.Windows.Forms;
using System.IO;

namespace TomTomCs
{
   public partial class Form1 : Form
   {
   
      private List<PDI> pdi = null;
   
      public Form1()
      {
         InitializeComponent();
      }

      private void btn_readPDI_Click(object sender, EventArgs e)
      {
         if (openFileDialog1.ShowDialog() == DialogResult.OK)
         {
            this.pdi = this.ReadTomTomPDI(openFileDialog1.FileName);
         }
      }

      private void btn_writePDI_Click(object sender, EventArgs e)
      {
         if (this.pdi != null && saveFileDialog1.ShowDialog() == DialogResult.OK)
         {
            this.WriteTomTomPDI(saveFileDialog1.FileName, this.pdi);
         }
      }

      #region PDI class
      public class PDI
      {
         public PDI() {}
         public double Longitude;
         public double Latitude;
         public string Description;
      }
      #endregion

      #region ReadTomTomPDI
      /// <summary>
      /// An OV2 file consists of a sequence of variable−length records. Every record starts with a one−byte
      /// "type". This type tells you how to process the rest of the record.
      ///
      /// You should not encounter types other than 0, 1, 2 and 3 − if you do, the file is either corrupt or in a
      /// different (e.g. a higher−version) format.
      ///
      /// Coordinates are stored as 4−byte integers representing a WGS84 longitude or latitude, multiplied by
      /// 100,000 and rounded to the nearest integer. As such, an X−coordinate should always be a value between
      /// −18,000,000 and +18,000,000, and a Y−coordinate should be a value between −9,000,000 and +9,000,000.
      ///
      /// DELETED RECORD:
      /// - 1 byte T: type (always 0)
      /// - 4 bytes L: length of this record in bytes (including the T and L fields)
      /// - L−5 bytes bytes to ignore (content undefined)
      ///
      /// SKIPPER RECORD:
      /// - 1 byte T: type (always 1)
      /// - 4 bytes Number of bytes in the file, including and starting at this
      /// record, that contain data for POI enclosed in the given rectangle.
      /// - 4 bytes X1: longitude coordinate of the west edge of the rectangle
      /// - 4 bytes Y1: latitude coordinate of the south edge of the rectangle
      /// - 4 bytes X2: longitude coordinate of the east edge of the rectangle
      /// - 4 bytes Y2: latitude coordinate of the north edge of the rectangle
      ///
      /// SIMPLE POI RECORD:
      /// - 1 byte T: type (always 2)
      /// - 4 bytes L: length of this record in bytes (including the T and L fields)
      /// - 4 bytes X: longitude coordinate of the POI
      /// - 4 bytes Y: latitude coordinate of the POI
      /// - L−13 bytes Name: zero−terminated ASCII string specifying the name of the POI
      ///
      /// EXTENDED POI RECORD:
      /// - 1 byte T: type (always 3)
      /// - 4 bytes L: length of this record in bytes (including the T and L fields)
      /// - 4 bytes X: longitude coordinate of the POI
      /// - 4 bytes Y: latitude coordinate of the POI
      /// - P bytes Name: zero−terminated ASCII string specifying the name of the POI
      /// - Q bytes Unique ID: zero−terminated string specifying the unique ID of the POI
      /// - L−P−Q−13 bytes Extra data: zero−terminated string, not used yet
      ///
      /// If you encounter any other type, either the file is corrupt, or the file contains extra (proprietary) records.
      /// In either case, you should stop processing immediately. Since there is always the danger that the file is
      /// corrupt, you should in fact wonder whether the preceding records read so far were in fact valid.
      /// </summary>
      /// <param name="path"></param>
      /// <returns></returns>
      public List<PDI> ReadTomTomPDI(string path)
      {
         List<PDI> pdi = new List<PDI>();
         try
         {
            if (File.Exists(path))
            {
               using (BinaryReader br = new BinaryReader(File.Open(path, FileMode.Open)))
               {
                  byte type;
                  int total;
                  while (br.BaseStream.Position < br.BaseStream.Length)
                  {
                     type = br.ReadByte();
                     switch (type)
                     {
                        case 0:    // DELETED RECORD
                           {
                              total = br.ReadInt32();
                              br.ReadBytes(total - 5);  // L−5 bytes bytes to ignore (content undefined)
                           }
                           break;

                        case 1:    // SKIPPER RECORD
                           {
                              total = br.ReadInt32();
                              double weLongitude = (double)((double)br.ReadInt32() / 100000.0);
                              double seLatitude = (double)((double)br.ReadInt32() / 100000.0);
                              double eeLongitude = (double)((double)br.ReadInt32() / 100000.0);
                              double neLatitude = (double)((double)br.ReadInt32() / 100000.0);
                           }
                           break;

                        case 2:    // SIMPLE POI RECORD
                           {
                              total = br.ReadInt32();
                              PDI pd = new PDI();
                              pd.Longitude = (double)((double)br.ReadInt32() / 100000.0);
                              pd.Latitude = (double)((double)br.ReadInt32() / 100000.0);
                              pd.Description = new string(br.ReadChars(total - 13)).Replace("\0", string.Empty);
                              pdi.Add(pd);
                           }
                           break;

                        case 3:    // EXTENDED POI RECORD
                           {
                              total = br.ReadInt32();
                              PDI pd = new PDI();
                              pd.Longitude = (double)((double)br.ReadInt32() / 100000.0);
                              pd.Latitude = (double)((double)br.ReadInt32() / 100000.0);
                              char c = ' ';
                              while ((c = br.ReadChar()) != '\0') {pd.Description += c;}
                              string uniqueID = string.Empty;
                              while ((c = br.ReadChar()) != '\0') { uniqueID += c; }
                              string extraData = string.Empty;
                              while ((c = br.ReadChar()) != '\0') { extraData += c; }
                              pdi.Add(pd);
                           }
                           break;
                           
                        default :
                           br.BaseStream.Position = br.BaseStream.Length-1;
                           break;
                     }
                  }
               }
            }            
         }
         catch (Exception)
         {
         }
         return pdi;
      }
      #endregion

      #region WriteTomTomPDI
      public void WriteTomTomPDI(string path, List<PDI> pdi)
      {
         try
         {
            using (BinaryWriter br = new BinaryWriter(File.Open(path, FileMode.Create)))
            {
               foreach(PDI pd in pdi)
               {
                  br.Write((byte)2);
                  br.Write((int)(pd.Description.Length + 13 + 1));  // +1 when add "/0" to description
                  br.Write((int)(pd.Longitude * 100000.0));
                  br.Write((int)(pd.Latitude * 100000.0));
                  br.Write((char[])pd.Description.ToCharArray());
                  br.Write((char)'\0');
               }
            }
         }
         catch (Exception)
         {
         }
      }
      #endregion

   }
}
